
/* U7PATCH.CPP
  -------------

  Program to run patch U7.EXE so that all accesses to voodoo memory are
  replaced by interrupt calls.  Removes the check for an EMS manager and
  forces the OS to report 2MB of extended memory.

  Copyright (C) 2000 by Josef Drexler <jdrexler@julian.uwo.ca>

  Distributed under the terms of the GNU General Public License.  For more
  information please read the file COPYING which is included in the
  distribution.

*/


#include <stdio.h>
#include <stdlib.h>
#include <grep.h>

#include "u7patch.h"

#define BLOCK 16384

char* block;
long opstart;

void check(int intnum)
{
//  printf("%4X", intnum);
  if ( (intnum < INT_FIRST) || (intnum > INT_LAST) ) {
	printf("!!!!\nABORTED!");
	exit(1);
  }
}

// calls one of the three-byte instruction interrupts
void putthree(int intnum, int intpar)
{
  check(intnum);
  block[opstart] = 0xcd;
  block[opstart+1] = intnum;
  block[opstart+2] = intpar;
}

// call one of the two-byte ints
void puttwo(int intnum)
{
  check(intnum);
  block[opstart] = 0xcd;
  block[opstart+1] = intnum;
}

char* mnemonic(int opcode)
{
  FILE *opcodes = fopen("opcodes.txt", "r");
  int  f_opcode;
  static char f_mnemonic[256];

  do {
	fscanf(opcodes, "%2X\t", &f_opcode);
	fgets(f_mnemonic, 256, opcodes);
  } while ( (opcode != f_opcode) && (f_opcode != 0x100) );
  fclose(opcodes);

  return f_mnemonic;
}

// finds the next occurence of 0x67
int nextmatch(char* block, long int last)
{
  char* found = (char*) memchr(block + last, 0x67, BLOCK - last);
  if (!found)
	return -1;
  return found - block;
}

// determines the index register from the postbyte
int getindex(int postbyte, int *memop, int *regop)
{
  if (postbyte == 0xb)
	printf("");

  int mode = postbyte >> 6;
  *regop = (postbyte >> 3) & 7;
  *memop = postbyte & 7;

  switch (mode) {
	case 0:
		if (*regop == 4)
			return 1;	// next byte is scale-base-index
		if (*regop == 5)
			return 2;	// next byte is ofs32
		return 0;
	case 1:
		return 3;	// [reg+ofs8]
	case 2:
		return 4;	// [reg+ofs32]
  }
  return 5;	// reg,reg
}

// special cases
void putspecial(char *block, long int offset)
{
  static int specialnum = 0;
  printf("Special %d is: %x %x %x %x %x...\n",
	specialnum, block[offset], block[offset+1], block[offset+2],
	block[offset+3], block[offset+4]);

  putthree(INT_SPCL, specialnum++);
}

// calls an int with "A" parameter style
int putparmsA(int intnum3, int intnum4, int opcode, int postbyte, char* opname)
{
  int index, memop, regop;

  index = getindex(postbyte, &memop, &regop);
  switch (index) {
	case 0:
		putthree(intnum3 + ((opcode >> 3) & 7),
			 (opcode & 3) << 6 | memop << 3 | regop);
		return 0;
	case 3:
		putthree(intnum4 + ((opcode >> 3) & 7),
			 (opcode & 3) << 6 | memop << 3 | regop);
		return 0;
	default:
		return 1;
  }
}

int main()
{
  FILE *u7exe = fopen("u7.exe", "r+b");
  char firstvoodoo[3] = { 0x67, 0x8b, 0x03 }; // mov ax,[ebx]
  long int voodoostart = grepfile(u7exe, firstvoodoo, sizeof firstvoodoo, 1);
  if (voodoostart == GREP_NOMATCHL) {
	printf("Signature not found\n");
	exit(1);
  }
  int totalops = 0, ignored = 0, unhandled = 0, future = 0, special = 0;

  fseek(u7exe, voodoostart, SEEK_SET);

  block = (char*) malloc(BLOCK);
  long int voodoo = 0;
  long int blocklen = fread(block, 1, BLOCK, u7exe);

	// get pointer to next voodoo instruction
  while ( (voodoo = nextmatch(block, voodoo)) >= 0) {
	// analyze instruction

	// see if there are prefixes following; if yes swap them to before
	int looking = 1;
	while (looking) {
		switch (block[voodoo + 1]) {
			case 0x66:
			case 0x26:
			case 0xf3:
				block[voodoo] = block[voodoo + 1];
				block[++voodoo] = 0x67;
				break;
			default: looking = 0;
		}
		break;
	}

	opstart = voodoo;
	int opcode = block[++voodoo];
	int postbyte, index, memop, regop;
	totalops++;

	switch (opcode >> 2) {
		case 0x22:
			postbyte = block[++voodoo];
			if (putparmsA(INT_MOV_3, INT_MOV_4, opcode & 3, postbyte, "MOV")) {
				putspecial(block, --voodoo);
				special++;
			}
			break;

		case 0x00:
		case 0x02:
		case 0x04:
		case 0x06:
		case 0x08:
		case 0x0a:
		case 0x0c:
		case 0x0e:
			postbyte = block[++voodoo];
			if (putparmsA(INT_OP1_3, INT_OP1_4, opcode, postbyte, "OP1")) {
				putspecial(block, --voodoo);
				special++;
			}
			break;

			// ignore these
		case 0x01:
		case 0x03:
		case 0x05:
		case 0x07:
		case 0x09:
		case 0x0b:
		case 0x0d:
		case 0x0f:
//			printf("Ignoring %s", mnemonic(opcode));
			ignored++;
			break;

			// handle these later
		case 0x29:	// MOVSx + CMPSx
		case 0x2a:	// TEST + STOSx
		case 0x2b:	// LODSx + SCASx
			puttwo(INT_SOP + opcode - 0xa4);
			break;
		case 0x31:	// MOV [mem],imm
			if (opcode < 0xc6) {
				putspecial(block, voodoo);
				special ++;
				break;
			}
			postbyte = block[++voodoo];
			if (putparmsA(INT_MOVIMM, 0, opcode & 1, postbyte, "MOV_imm")) {
				putspecial(block, --voodoo);
				special ++;
			}
			break;
		case 0x35:	// XLAT
			puttwo(INT_XLAT);
			break;
		case 0x3d:	// OP2 [mem]{,imm}
			putspecial(block, voodoo);
			special++;
			break;

		default:
			printf("Don't know how to handle opcode %x at %lx: %s",
				opcode, voodoo, mnemonic(opcode));
			unhandled++;
			break;
	}
  }
  fseek(u7exe, voodoostart, SEEK_SET);
  fwrite(block, 1, blocklen, u7exe);

  printf("Out of %d instructions, %d were handled, %d were ignored \n"
	"%d are special cases, %d will be handled and %d were not handled\n"
	"Interrupt range is %x-%x\n",
	totalops, totalops - ignored - unhandled - future - special,
	ignored, special, future, unhandled, INT_FIRST, INT_LAST);

  printf("\nRemoving EMMXXXX0 check: ");
  fseek(u7exe, 0, SEEK_SET);
  long int emmxxxx0 = grepfile(u7exe, "EMMXXXX0", 8, 1);

  if (emmxxxx0 == GREP_NOMATCHL) {
	printf("Not found?");
  } else {
	fseek(u7exe, emmxxxx0, SEEK_SET);
	fwrite("UUU", 1, 3, u7exe);
	printf("Done");
  }

  printf("\nRemoving extended memory size checks:");
  fseek(u7exe, 0, SEEK_SET);
  for (int occurance=0; occurance<2; occurance++) {
	char oldcode[4] = { 0xb4, 0x88, 0xcd, 0x15 };	// mov ah,88;int 15
	long int memcheck = grepfile(u7exe, "", sizeof oldcode, 1);
	printf(" %d/2 ", occurance+1);
	if (memcheck == GREP_NOMATCHL) {
		printf("Not found?");
		break;
	} else {
		char newcode[4] = { 0xb8, 0, VOODOOMBS, 0x90 }; // mov ax,VOODOOMBS
		fseek(u7exe, memcheck, SEEK_SET);
		fwrite(newcode, 1, sizeof newcode, u7exe);
		printf("Done");
	}
  }

  printf("\n\nAll Done! Now try running U7RUN.EXE\n");

  fclose(u7exe);

  return 0;
}
